program pipc(input, output, stderr);

{$mode objfpc}

uses
	ipc,
	// baseUnix for error numbers and fpGetErrno()
	baseUnix,
	// for pError() [print error]
	errors,
	// gettext for internationalization [translateResourcestrings()]
	gettext,
	// sysutils for applicationName() and format()
	sysutils;

const
	seed = '/etc/fstab';
	
	exitCodeFileToKeyError = 1;
	exitCodeMsgGetError    = 2;
	exitCodeMsgStatError   = 3;
	exitCodeMsgSndError    = 4;
	exitCodeMsgRmError     = 5;

// resource strings {{{
resourcestring
	fileToKeyError = '☠: ¿🔑 ''%s''? ⇒ ¬🔑';
	fileToKeyErrorAccess = 'ftok(''%s'', id) ⇒ errno = EACCES';
	fileToKeyErrorBadAddress = 'ftok() ⇒ errno = EBADF';
	fileToKeyErrorLoop = 'ftok(''%s'', id) ⇒ errno = ELOOP';
	fileToKeyErrorPathTooLong = 'ftok(''%s'', id) ⇒ errno = ENAMETOOLONG';
	fileToKeyErrorNonexistent = 'ftok(''%s'', id) ⇒ errno = ENOENT';
	fileToKeyErrorOOM = 'ftok() ⇒ errno = ENOMEM';
	fileToKeyErrorNotDir = 'ftok(''%s'', id) ⇒ errno = ENOTDIR';
	fileToKeyErrorOverflow = 'ftok() ⇒ errno = EOVERFLOW';
	
	msgGetError = '☠: ¿🖅 “%s”? ⇒ ¬🖅';
	msgGetErrorAccess = 'msgget(“%s”, flag) ⇒ errno = EACCES';
	msgGetErrorNoMem = 'msgget() ⇒ errno = ENOMEM';
	msgGetErrorExceedingMsgQueueMax = 'msgget() ⇒ errno = ENOSPC';
	
	msgStatError = '☠: ¿ℹ “%s” 🖅? ⇒ ¬ℹ';
	msgStatErrorAccess = 'msgctl(“%s”, IPC_STAT, buf) ⇒ errno = EACCES';
	msgStatErrorFault = 'msgctl() ⇒ errno = EFAULT';
	msgStatErrorQueueRemoved = 'msgctl(“%s”, IPC_STAT, buf) ⇒ errno = EIDRM';
	msgStatErrorInvalidParameter = 'msgctl() ⇒ errno = EINVAL';
	
	msgSndError = '☠: ¡ℹ “%s” 📫! ↯';
	msgSndErrorAccess = 'msgsnd(“%s”, …) ⇒ errno = EACCES';
	msgSndErrorAgain = 'msgsnd() ⇒ errno = EAGAIN';
	msgSndErrorFault = 'msgsnd() ⇒ errno = EFAULT';
	msgSndErrorQueueRemoved = 'msgsnd(“%s”, …) ⇒ errno = EIDRM';
	msgSndErrorInterrupt = 'msgsnd() ⇒ errno = EINTR';
	msgSndErrorInvalidParameter = 'msgsnd() ⇒ errno = EINVAL';
	msgSndErrorNoMem = 'msgsnd() ⇒ errno = ENOMEM';
	
	msgRmError = '☠: ¡ℹ “%s” ♺! ↯';
	msgRmErrorQueueRemoved = 'msgctl(“%s”, IPC_IDRM, nil) ⇒ errno = EIDRM';
	msgRmErrorInvalidParameter = 'msgctl() ⇒ errno = EINVAL';
	msgRmErrorPermission = 'msgctl(“%s”, IPC_IDRM, nil) ⇒ errno = EPERM';
// }}}

type
	// IPC_perm {{{
	// struct ipc_perm {…} found in /usr/include/x86_64-linux-gnu/bits/ipc.h
	/// \brief data structure used to
	///        pass permission information to IPC operations
	//TIPC_perm = record
	//	/// \brief key
	//	key: TKey;
	//	/// \brief owner's user ID
	//	uid: kernel_uid_t;
	//	/// \brief owner's group ID
	//	gid: kernel_gid_t;
	//	/// \brief creator's user ID
	//	cuid: kernel_uid_t;
	//	/// \brief creator's group ID
	//	cgid: kernel_gid_t;
	//	/// \brief read/write permission
	//	mode: kernel_mode_t;
	//	pad0: cushort;
	//	/// \brief sequence number
	//	seq: cushort;
	//	pad1: cushort;
	//	glibc_reserved0: culong;
	//	glibc_reserved1: culong;
	//end;
	// }}}
	
	// MsqId_DS {{{
	// struct msqid_ds {…} found in /usr/include/x86_64-linux-gnu/bits/msq.h
	/// \brief Structure of record for one message inside the kernel.
	/// 
	/// The type `struct msg` is opaque.
	TMSQid_ds = record
		/// \brief structure describing operation permission
		msg_perm: TIPC_perm;
		/// \brief time of last msgsnd command
		msg_stime: time_t;
		{$ifndef CPUX86_64}
			glibc_reserved0: culong;
		{$endif}
		/// \brief time of last msgrcv command
		msg_rtime: time_t;
		{$ifndef CPUX86_64}
			glibc_reserved1: culong;
		{$endif}
		/// \brief time of last change
		msg_ctime: time_t;
		{$ifndef CPUX86_64}
			glibc_reserved2: culong;
		{$endif}
		/// \brief current number of bytes on queue
		msg_cbytes: msglen_t;
		/// \brief number of messages currently on queue
		msg_qnum: msgqnum_t;
		/// \brief max number of bytes allowed on queue
		msg_qbytes: msglen_t;
		/// \brief pid of last msgsnd()
		msg_lspid: pid_t;
		/// \brief pid of last msgrcv()
		msg_lrpid: pid_t;
		glibc_reserved3: culong;
		glibc_reserved4: culong;
	end;
	// }}}

var
	/// \brief key as retrieved via ftok(3)
	key: TKey;
	/// \brief message queue identifier retrieved via msgget(2)
	messageQueue: cint;
	/// \brief contains all data about our message queue
	queueInfo: TMSQid_ds;
	/// \brief temporarily stores return values
	functionReturnValue: cint;
	/// \brief contains a single message we wanna send via IPC
	msg:{$warning variable 'msg' uses IPC.TMsgBuf}TMsgBuf;
	/// \brief temporary iterator variable
	i: longint;

begin
	translateResourcestrings(applicationName() + '.%s.mo');
	
	// ftok(path, id): id only eight LSBs significant and
	//                 behavior with 0 (zero) unspecified
	key := fToK(seed, hash(applicationName()) mod 255 + 1);
	// fToK() error treatment {{{
	if key = -1 then
	begin
		writeLn(stderr, format(fileToKeyError, [seed]));
		case fpGetErrno() of
			ESysEAcces:
			begin
				writeLn(stderr, format(fileToKeyErrorAccess, [seed]));
			end;
			ESysEFault:
			begin
				writeLn(stderr, fileToKeyErrorBadAddress);
			end;
			ESysELoop:
			begin
				writeLn(stderr, format(fileToKeyErrorLoop, [seed]));
			end;
			ESysENameTooLong:
			begin
				writeLn(stderr, format(fileToKeyErrorPathTooLong, [seed]));
			end;
			ESysENoEnt:
			begin
				writeLn(stderr, format(fileToKeyErrorNonexistent, [seed]));
			end;
			ESysENoMem:
			begin
				writeLn(stderr, fileToKeyErrorOOM);
			end;
			ESysENotDir:
			begin
				writeLn(stderr, format(fileToKeyErrorNotDir, [seed]));
			end;
			ESysEOverflow:
			begin
				writeLn(stderr, fileToKeyErrorOverflow);
			end;
			else
			begin
				pError('ftok()', fpGetErrno());
			end;
		end;
		halt(exitCodeFileToKeyError);
	end;
	// }}}
	
	messageQueue := msgget(key, IPC_CREAT or &666);
	// msgget() error treatment {{{
	if messageQueue = -1 then
	begin
		writeLn(stderr, format(msgGetError, [seed]));
		case fpGetErrno() of
			ESysEAcces:
			begin
				writeLn(stderr, format(msgGetErrorAccess, [seed]));
			end;
			ESysENoMem:
			begin
				writeLn(stderr, msgGetErrorNoMem);
			end;
			ESysENoSpc:
			begin
				writeLn(stderr, msgGetErrorExceedingMsgQueueMax);
			end;
			else
			begin
				pError('msgget()', fpGetErrno());
			end;
		end;
		halt(exitCodeMsgGetError);
	end;
	// }}}
	
	for i := 0 to 5 do
	begin
		// check, if the queue has data
		functionReturnValue := msgctl(messageQueue, IPC_STAT, @queueInfo);
		// msgctl() error treatment {{{
		if functionReturnValue = -1 then
		begin
			writeLn(stderr, msgStatError);
			case fpGetErrno() of
				ESysEAcces:
				begin
					writeLn(stderr, format(msgStatErrorAccess, [seed]));
				end;
				ESysEFault:
				begin
					writeLn(stderr, msgStatErrorFault);
				end;
				ESysEIDRm:
				begin
					writeLn(stderr, format(msgStatErrorQueueRemoved, [seed]));
				end;
				ESysEInval:
				begin
					writeLn(stderr, msgStatErrorInvalidParameter);
				end;
				else
				begin
					pError('msgctl()', fpGetErrno());
				end;
			end;
			halt(exitCodeMsgStatError);
		end;
		// }}}
		
		writeLn('msg_qnum:   ', queueInfo.msg_qnum);
		
		if queueInfo.msg_qnum < 1 then
		begin
			msg.mtype := 343;
			msg.mtext[0] := '1';
			
			functionReturnValue := msgsnd(messageQueue, @msg, sizeOf(msg.mtext), 0);
			// msgsnd error treatment {{{
			if functionReturnValue = -1 then
			begin
				writeLn(stderr, msgSndError);
				case fpGetErrno() of
					ESysEAcces:
					begin
						writeLn(stderr, format(msgSndErrorAccess, [seed]));
					end;
					ESysEAgain:
					begin
						writeLn(stderr, msgSndErrorAgain);
					end;
					ESysEFault:
					begin
						writeLn(stderr, msgSndErrorFault);
					end;
					ESysEIDRm:
					begin
						writeLn(stderr, format(msgSndErrorQueueRemoved, [seed]));
					end;
					ESysEIntr:
					begin
						writeLn(stderr, msgSndErrorInterrupt);
					end;
					ESysEInval:
					begin
						writeLn(stderr, msgSndErrorInvalidParameter);
					end;
					ESysENoMem:
					begin
						writeLn(stderr, msgSndErrorNoMem);
					end;
					else
					begin
						pError('msgsnd()', fpGetErrno());
					end;
				end;
				halt(exitCodeMsgSndError);
			end;
			// }}}
		end;
		
		fpSleep(1);
	end;
	
	functionReturnValue := msgCtl(messageQueue, IPC_RMID, nil);
	// msgCtl() error treatment {{{
	if functionReturnValue = -1 then
	begin
		writeLn(stderr, msgRmError);
		case fpGetErrno() of
			ESysEIDRm:
			begin
				writeLn(stderr, format(msgRmErrorQueueRemoved, [seed]));
			end;
			ESysEInval:
			begin
				writeLn(stderr, msgRmErrorInvalidParameter);
			end;
			ESysEPerm:
			begin
				writeLn(stderr, format(msgRmErrorPermission, [seed]));
			end;
			else
			begin
				pError('msgctl()', fpGetErrno());
			end;
		end;
		halt(exitCodeMsgRmError);
	end;
	// }}}
	
	exitCode := 0;
end.
// vim: set tw=78 ts=4 noet bs=eol,start,indent fdm=marker fdc=1:
